#include <Arduino.h>

// wifi
#include <WiFi.h>
#include <HTTPClient.h>
#include <ArduinoJson.h>

// Настройки
#include "settings.h"

// ntp
#include <GyverNTP.h>

// epd
#include "epd_highlevel.h"
#include "epdiy.h"

// battery
#include <driver/adc.h>
#include "esp_adc_cal.h"
#include "esp_sleep.h"

// fonts
#include "IBMPlexMono12.h"
#include "MartianMono30.h"
#include "MartianMono200.h"

#define BATT_PIN 36
#define EPD_WIDTH 960
#define EPD_HEIGHT 540
#define WAVEFORM EPD_BUILTIN_WAVEFORM

const String MONTHS[12] = {"Января", "Февраля", "Марта", "Апреля", "Мая", "Июня", "Июля", "Августа", "Сентября", "Октября", "Ноября", "Декабря"};
const String WEEK_DAY[7] = {"Понедельник", "Вторник", "Среда", "Четверг", "Пятница", "Суббота", "Воскресенье"};

const uint8_t black = 0x0;

/**
 * Распиновка https://ae01.alicdn.com/kf/H133488d889bd4dd4942fbc1415e0deb1j.jpg
 */
gpio_num_t FIRST_BTN_PIN = GPIO_NUM_39;

EpdiyHighlevelState hl;

uint8_t *fb;           // Фрейм буфер экрана
int temperature = 20;  // Температура экрана
enum EpdDrawError err; // Возможная ошибка экрана

GyverNTP ntp(UTC_HOUR);                     // NTP Дата
EpdRotation orientation = EPD_ROT_PORTRAIT; // Ориентация экрана

int vref = 1100; // Вольтаж аккумулятора

/**
 * RTC_DATA_ATTR Данные которые сохраняются во время сна
 **/
RTC_DATA_ATTR int pressed_wakeup_btn_index; // Кол-во нажатий на кнопку
RTC_DATA_ATTR int current_day;              // Текущий день
RTC_DATA_ATTR float current_battery;        // Текущий день
RTC_DATA_ATTR String current_temp_update;   // Текущий день

/**
 * ? А оно нам надо может сразу спать пойдем?
 * That's maximum 10 seconds of runtime in microseconds
 */
int64_t maxTimeRunning = 10000000;

/**
 * Получение уровня заряда батареи в процентах
 */
double_t get_battery_percentage()
{
    // When reading the battery voltage, POWER_EN must be turned on
    epd_poweron();
    delay(50);

    Serial.println("Температура контроллера: " + String(epd_ambient_temperature()) + " °C");

    uint16_t v = analogRead(BATT_PIN);
    Serial.print("Аналоговое значения батареи: ");
    Serial.println(v);
    double_t battery_voltage = ((double_t)v / 4095.0) * 2.0 * 3.3 * (vref / 1000.0);
    double_t percent_experiment = ((battery_voltage - 3.7) / 0.5) * 100;
    if (percent_experiment > 100)
    {
        percent_experiment = 100;
    }

    String voltage = "Вольтаж батареи: " + String(battery_voltage) + "V примерно " + String(percent_experiment) + "%";
    Serial.println(voltage);

    return percent_experiment;
}

/**
 * Ложимся спать и указываем как будем просыпаться
 */
void start_deep_sleep_with_wakeup_sources()
{
    epd_poweroff(); // Выключаем дисплей
    delay(400);
    esp_sleep_enable_timer_wakeup(TIME_TO_SLEEP * uS_TO_S_FACTOR); // Каждые 5 минут просыпаемся
    esp_sleep_enable_ext0_wakeup(FIRST_BTN_PIN, 0);                // Включаем первую кнопку для пробуждения

    Serial.println("Отправляемся спать zzz");
    esp_deep_sleep_start();
}

/**
 * Пишем в serial от чего проснулись
 */
void print_wakeup_reason()
{
    esp_sleep_wakeup_cause_t wakeup_reason;
    wakeup_reason = esp_sleep_get_wakeup_cause();
    switch (wakeup_reason)
    {
    case ESP_SLEEP_WAKEUP_EXT0:
        Serial.println("Проснулись от внешнего источника RTC_IO");
        break;
    case ESP_SLEEP_WAKEUP_EXT1:
        Serial.println("Проснулись от внешнего источника RTC_CNTL");
        break;
    case ESP_SLEEP_WAKEUP_TIMER:
        Serial.println("Проснулись по таймеру");
        break;
    case ESP_SLEEP_WAKEUP_TOUCHPAD:
        Serial.println("Проснулись от касания");
        break;
    case ESP_SLEEP_WAKEUP_ULP:
        Serial.println("Проснулись от датчика");
        break;
    default:
        Serial.printf("Не понятно от чего проснулись: %d\n", wakeup_reason);
        break;
    }
}

/**
 * Поправить опорное напряжение АЦП
 */
void correct_adc_reference()
{
    esp_adc_cal_characteristics_t adc_chars;
    esp_adc_cal_value_t val_type = esp_adc_cal_characterize(ADC_UNIT_1, ADC_ATTEN_DB_11, ADC_WIDTH_BIT_12, 1100, &adc_chars);
    if (val_type == ESP_ADC_CAL_VAL_EFUSE_VREF)
    {
        Serial.printf("eFuse Vref:%u mV", adc_chars.vref);
        vref = adc_chars.vref;
    }
}

/**
 * Инициализация дисплея
 */
void epd_setup()
{
    epd_init(&epd_board_lilygo_t5_47, &ED097TC2, EPD_LUT_64K);
    hl = epd_hl_init(WAVEFORM);
    epd_set_rotation(orientation);
    fb = epd_hl_get_framebuffer(&hl);
    // epd_clear();
}

/**
 * Отображаем новые день и месяц
 */
void display_new_day()
{
    // Сбрасываем дисплей
    epd_clear();

    // Сбрасываем буфер
    epd_hl_set_all_white(&hl);

    // Курсор где начинается текст
    // ?? Хрень какая то с ним надо по другому придумать
    int cursor_x = (epd_height() / 2) - 170;
    int cursor_y = (epd_width() / 2) + 150;

    String day;
    day += ntp.day();
    String day_week = WEEK_DAY[ntp.dayWeek() - 1];
    String month = MONTHS[ntp.month() - 1];

    EpdFontProperties font_props = epd_font_properties_default();
    font_props.flags = EPD_DRAW_ALIGN_CENTER;
    epd_write_string(&MartianMono200, day.c_str(), &cursor_x, &cursor_y, fb, &font_props);
    cursor_y += 100;
    cursor_x -= 160;
    epd_write_string(&MartianMono30, month.c_str(), &cursor_x, &cursor_y, fb, &font_props);

    cursor_x -= 160;
    cursor_y += 30;
    epd_write_string(&MartianMono30, day_week.c_str(), &cursor_x, &cursor_y, fb, &font_props);

    Serial.println("X: " + String(cursor_x));
    Serial.println("Y: " + String(cursor_y));

    epd_poweron();
    while (epd_hl_update_screen(&hl, MODE_GC16, temperature) != EPD_DRAW_SUCCESS)
    {
        Serial.println("Ошибка в отрисовки нового дня");
        delay(600);
    };
    epd_poweroff();
}

/**
 * Отрисовка погоды
 */
String httpGETRequest(const char *URL)
{
    WiFiClient client;
    HTTPClient http;
    http.addHeader("Authorization", HA_TOKEN);

    http.begin(client, URL);
    // Send HTTP POST request
    int httpResponseCode = http.GET();

    String payload = "{}";

    if (httpResponseCode > 0)
    {
        Serial.print("HTTP Response code: ");
        Serial.println(httpResponseCode);
        payload = http.getString();
    }
    else
    {
        Serial.print("Error code: ");
        Serial.println(httpResponseCode);
    }
    // Free resources
    http.end();
    return payload;
}


/**
 * Отчистка части экрана
 */
void clear_area(EpdRect area)
{
    epd_poweron();
    epd_clear_area(area);
    epd_poweroff();
    delay(300);
}

/**
 * Отрисовка погоды
 */
void draw_weather()
{
    // Запрос в HA
    String httpReply = httpGETRequest(HA_ENDPOINT);
    // Allocate the JSON document
    JsonDocument doc;
    // Deserialize the JSON document
    DeserializationError error = deserializeJson(doc, httpReply.c_str());
    // Test if parsing succeeds
    if (error)
    {
        Serial.print(F("deserializeJson() failed: "));
        Serial.println(error.f_str());
    }
    else
    {
        String last_updated = doc["last_updated"];

        if (current_temp_update == last_updated)
        {
            Serial.println("Погода не изменилась");
        }
        else
        {
            current_temp_update = last_updated;
            EpdRect info_area = {
                .x = 150,
                .y = 0,
                .width = 100,
                .height = EPD_HEIGHT,
            };
            clear_area(info_area);

            String temperature_line = doc["attributes"]["yandex_condition"];
            long temperature = doc["attributes"]["temperature"];
            temperature_line.concat(" ");
            temperature_line += String(temperature);
            temperature_line.concat("°C");

            int temp_cursor_x = EPD_HEIGHT / 2;
            int temp_cursor_y = EPD_WIDTH + 30;

            EpdFontProperties temp_font_props = epd_font_properties_default();
            temp_font_props.flags = EPD_DRAW_ALIGN_CENTER;

            temperature = doc["attributes"]["feels_like"];
            temperature_line += "\nОщущается как ";
            temperature_line += String(temperature);
            temperature_line.concat("°C");

            epd_write_string(&IBMPlexMono12, temperature_line.c_str(), &temp_cursor_x, &temp_cursor_y, fb, &temp_font_props);

            epd_poweron();
            // err = epd_hl_update_area(&hl, MODE_DU, temperature, info_area);
            err = epd_hl_update_screen(&hl, MODE_DU, temperature);
            epd_poweroff();
        }
    }
}

/**
 * Отрисовка заряда акб
 */
void draw_battery()
{

    double_t battery_percentage = get_battery_percentage();
    if (battery_percentage == current_battery)
    {
        return;
    }
    else
    {
        EpdRect info_area = {
            .x = 0,
            .y = 0,
            .width = 100,
            .height = EPD_HEIGHT,
        };
        clear_area(info_area);
        current_battery = battery_percentage;

        int battery_cursor_x = EPD_HEIGHT - 30;
        int battery_cursor_y = EPD_WIDTH + 220;

        EpdFontProperties battery_font_props = epd_font_properties_default();
        battery_font_props.flags = EPD_DRAW_ALIGN_RIGHT;
        String battery_text = String(battery_percentage);
        battery_text.concat(" %");
        epd_write_string(&IBMPlexMono12, battery_text.c_str(), &battery_cursor_x, &battery_cursor_y, fb, &battery_font_props);

        epd_poweron();
        // err = epd_hl_update_area(&hl, MODE_DU, temperature, info_area);
        err = epd_hl_update_screen(&hl, MODE_DU, temperature);
        epd_poweroff();
    }
}


/**
 * Установка
 */
void setup()
{
    Serial.begin(115200);
    print_wakeup_reason();
    correct_adc_reference();

    epd_setup(); // Установки дисплея

    WiFi.begin(ESP_WIFI_SSID, ESP_WIFI_PASSWORD);
    while (WiFi.status() != WL_CONNECTED)
    {
        Serial.println("Подключаемся к wifi...");
        delay(100);
    };
    Serial.println("Подключились к wifi сети");

    ntp.begin();
    while (ntp.updateNow() != 0)
    {
        Serial.println("Подключаемся к ntp серверу");
        delay(100);
    }
    Serial.println("Подключились к ntp серверу");
    Serial.println("Текущее время: " + ntp.timeString() + " " + ntp.dateString());

    esp_sleep_wakeup_cause_t wakeup_reason = esp_sleep_get_wakeup_cause();
    if (wakeup_reason == ESP_SLEEP_WAKEUP_EXT0)
    {
        Serial.println("Проснулись от нажатия кнопки");
        pressed_wakeup_btn_index++;
    }
    else
    {
        Serial.println("Проснулись от кнопкой сброса или циклом включения");
    }

    // Обновляем календарь если наступил новый день.
    if (current_day != ntp.day())
    {
        current_day = ntp.day();
        Serial.println("Переворачиваем календарь");
        display_new_day(); // Обновляем экран целиком
    }
    else
    {
        Serial.println("Новый день не наступил");
    }

    draw_weather(); // Обновляем информацию о погоде
    draw_battery(); // Обновляем заряд батареи

    Serial.println("Все отрисовали.");
}

/**
 * Цикл программы
 */
void loop()
{
    ntp.tick();
    if (esp_timer_get_time() > maxTimeRunning)
    {
        start_deep_sleep_with_wakeup_sources();
    }
}